@@ -23,8 +23,6 @@ from owlready2.util import ContextVar
 from owlready2 import *
 
 
-class _List(list): pass
-
 lg = rply.LexerGenerator()
 lg.add("}",                    r"\}")
 lg.add("{",                    r"\{")
@@ -40,6 +38,7 @@ lg.add("DECIMAL",              r"""[+-]?[0-9]*\.[0-9]+""")
 lg.add("INTEGER",              r"""[+-]?[0-9]+""")
 lg.add("-",                    r"-")
 lg.add("+",                    r"\+")
+lg.add("*STATIC",              r"\*STATIC")
 lg.add("*",                    r"\*")
 lg.add("/",                    r"/")
 lg.add(",",                    r",")
@@ -60,7 +59,8 @@ lg.add("?",                    r"""\?""")
 lg.add(".",                    r"""\.""")
 lg.add("PREFIXED_NAME",        r"""[\w\.\-]*:([\w\.\-:]|%[0-9a-fA-F]{2}|\\[_~\.\-!$&"'()*+,;=/?#@%])*([\w\-:]|%[0-9a-fA-F]{2}|\\[_~\.\-!$&"'()*+,;=/?#@%])""")
 lg.add("PNAME_NS",             r"""[\w\.\-]*:""")
-lg.add("FUNC",                 r"""(?:STRLANG)|(?:STRDT)|(?:STRLEN)|(?:STRSTARTS)|(?:STRENDS)|(?:STRBEFORE)|(?:STRAFTER)|(?:LANGMATCHES)|(?:LANG)|(?:DATATYPE)|(?:BOUND)|(?:IRI)|(?:URI)|(?:BNODE)|(?:RAND)|(?:ABS)|(?:CEIL)|(?:FLOOR)|(?:ROUND)|(?:CONCAT)|(?:STR)|(?:UCASE)|(?:LCASE)|(?:ENCODE_FOR_URI)|(?:CONTAINS)|(?:YEAR)|(?:MONTH)|(?:DAY)|(?:HOURS)|(?:MINUTES)|(?:SECONDS)|(?:TIMEZONE)|(?:TZ)|(?:NOW)|(?:UUID)|(?:STRUUID)|(?:MD5)|(?:SHA1)|(?:SHA256)|(?:SHA384)|(?:SHA512)|(?:COALESCE)|(?:IF)|(?:sameTerm)|(?:isIRI)|(?:isURI)|(?:isBLANK)|(?:isLITERAL)|(?:isNUMERIC)|(?:REGEX)|(?:SUBSTR)|(?:REPLACE)|(?:SIMPLEREPLACE)|(?:NEWINSTANCEIRI)\b""", re.IGNORECASE)
+lg.add("FUNC",                 r"""(?:STRLANG)|(?:STRDT)|(?:STRLEN)|(?:STRSTARTS)|(?:STRENDS)|(?:STRBEFORE)|(?:STRAFTER)|(?:LANGMATCHES)|(?:LANG)|(?:DATATYPE)|(?:BOUND)|(?:IRI)|(?:URI)|(?:BNODE)|(?:RAND)|(?:ABS)|(?:CEIL)|(?:FLOOR)|(?:ROUND)|(?:CONCAT)|(?:STR)|(?:UCASE)|(?:LCASE)|(?:ENCODE_FOR_URI)|(?:CONTAINS)|(?:YEAR)|(?:MONTH)|(?:DAY)|(?:HOURS)|(?:MINUTES)|(?:SECONDS)|(?:TIMEZONE)|(?:TZ)|(?:NOW)|(?:UUID)|(?:STRUUID)|(?:MD5)|(?:SHA1)|(?:SHA256)|(?:SHA384)|(?:SHA512)|(?:COALESCE)|(?:IF)|(?:sameTerm)|(?:isIRI)|(?:isURI)|(?:isBLANK)|(?:isLITERAL)|(?:isNUMERIC)|(?:REGEX)|(?:SUBSTR)|(?:REPLACE)|(?:SIMPLEREPLACE)|(?:NEWINSTANCEIRI)|(?:LOADED)|(?:STORID)|(?:DATETIME)|(?:DATE)|(?:TIME)\b""", re.IGNORECASE)
+#lg.add("FUNC",                 r"""(?:STRLANG)|(?:STRDT)|(?:STRLEN)|(?:STRSTARTS)|(?:STRENDS)|(?:STRBEFORE)|(?:STRAFTER)|(?:LANGMATCHES)|(?:LANG)|(?:DATATYPE)|(?:BOUND)|(?:IRI)|(?:URI)|(?:BNODE)|(?:RAND)|(?:ABS)|(?:CEIL)|(?:FLOOR)|(?:ROUND)|(?:CONCAT)|(?:STR)|(?:UCASE)|(?:LCASE)|(?:ENCODE_FOR_URI)|(?:CONTAINS)|(?:YEAR)|(?:MONTH)|(?:DAY)|(?:HOURS)|(?:MINUTES)|(?:SECONDS)|(?:TIMEZONE)|(?:TZ)|(?:NOW)|(?:UUID)|(?:STRUUID)|(?:MD5)|(?:SHA1)|(?:SHA256)|(?:SHA384)|(?:SHA512)|(?:COALESCE)|(?:IF)|(?:sameTerm)|(?:isIRI)|(?:isURI)|(?:isBLANK)|(?:isLITERAL)|(?:isNUMERIC)|(?:REGEX)|(?:SUBSTR)|(?:REPLACE)|(?:SIMPLEREPLACE)|(?:NEWINSTANCEIRI)|(?:XSD:DOUBLE)|(?:XSD:INTEGER)\b""", re.IGNORECASE)
 lg.add("MINUS",                r"""MINUS\b""", re.IGNORECASE)
 lg.add("AGGREGATE_FUNC",       r"""(?:COUNT)|(?:SUM)|(?:MIN)|(?:MAX)|(?:AVG)|(?:SAMPLE)|(?:GROUP_CONCAT)\b""", re.IGNORECASE)
 lg.add("BASE",                 r"""BASE\b""", re.IGNORECASE)
@@ -72,8 +72,8 @@ lg.add("EXISTS",               r"""EXISTS\b""", re.IGNORECASE)
 lg.add("NOT_EXISTS",           r"""NOT\s+EXISTS\b""", re.IGNORECASE)
 lg.add("FILTER",               r"""FILTER\b""", re.IGNORECASE)
 lg.add("UNION",                r"""UNION\b""", re.IGNORECASE)
-#lg.add("UNDEF",                r"""UNDEF\b""", re.IGNORECASE)
-#lg.add("VALUES",               r"""VALUES\b""", re.IGNORECASE)
+lg.add("UNDEF",                r"""UNDEF\b""", re.IGNORECASE)
+lg.add("VALUES",               r"""VALUES\b""", re.IGNORECASE)
 lg.add("BIND",                 r"""BIND\b""", re.IGNORECASE)
 lg.add("WITH",                 r"""WITH\b""", re.IGNORECASE)
 lg.add("AS",                   r"""AS\b""", re.IGNORECASE)
@@ -107,6 +107,7 @@ lg.add("ASC",                  r"""ASC\b""", re.IGNORECASE)
 lg.add("DESC",                 r"""DESC\b""", re.IGNORECASE)
 lg.add("LIMIT",                r"""LIMIT\b""", re.IGNORECASE)
 lg.add("OFFSET",               r"""OFFSET\b""", re.IGNORECASE)
+lg.add("STATIC",               r"STATIC")
 #lg.add("DESCRIBE",             r"""DESCRIBE\b""", re.IGNORECASE)
 #lg.add("ASK",                  r"""ASK\b""", re.IGNORECASE)
 #lg.add("CONSTRUCT",            r"""CONSTRUCT\b""", re.IGNORECASE)
@@ -134,6 +135,8 @@ def f(p): return _parse_select_query(p[1])
 def _parse_select_query(p):
   translator = CURRENT_TRANSLATOR.get()
   if isinstance(p[2], rply.Token) and (p[2].value == "*"): p[2] = None
+  if isinstance(p[4], UnionBlock) and (p[5][0] or p[5][1] or p[5][2] or p[5][3] or p[5][4]): # UNION with some solution modifier
+    p[4] = SimpleTripleBlock([p[4]])
   main_query = translator.new_sql_query("main", p[4], p[2], p[1], p[5])
   main_query.type = "select"
   return main_query
@@ -143,7 +146,7 @@ def _parse_select_query(p):
 def f(p): return p[1]
 #pg.optional(";?")
 
-@pg.production("prefix_decl : PREFIX PNAME_NS iri")
+@pg.production("prefix_decl : PREFIX PNAME_NS unabbreviated_iri")
 def f(p):
   CURRENT_TRANSLATOR.get().prefixes[p[1].value] = p[2].value
   return None
@@ -317,7 +320,7 @@ def _rename_params(d, x):
   elif isinstance(x, rply.Token) and (x.name == "PARAM"):
     x.number = d[x.number]
     
-def _create_modify_query(ontology_iri, deletes, inserts, using, group_graph_pattern):
+def _create_modify_query(ontology_iri, deletes, inserts, using, group_graph_pattern, solution_modifier):
   translator = CURRENT_TRANSLATOR.get()
   selects = []
   vars    = set()
@@ -333,7 +336,7 @@ def _create_modify_query(ontology_iri, deletes, inserts, using, group_graph_patt
   old_2_new_param = { n : i+1 for i, n in enumerate(select_param_indexes) }
   _rename_params(old_2_new_param, group_graph_pattern)
   
-  main_query = translator.new_sql_query("main", group_graph_pattern, selects, None, [None, None, None, None, None])
+  main_query = translator.new_sql_query("main", group_graph_pattern, selects, None, solution_modifier)
   main_query.type                 = "modify"
   main_query.ontology_iri         = ontology_iri
   main_query.inserts              = inserts
@@ -341,12 +344,12 @@ def _create_modify_query(ontology_iri, deletes, inserts, using, group_graph_patt
   main_query.select_param_indexes = select_param_indexes
   return main_query
 
-@pg.production("modify : with_iri? insert_clause using_clause*  WHERE group_graph_pattern")
-def f(p): return _create_modify_query(p[0], [], p[1], p[2], p[4])
-@pg.production("modify : with_iri? delete_clause using_clause*  WHERE group_graph_pattern")
-def f(p): return _create_modify_query(p[0], p[1], [], p[2], p[4])
-@pg.production("modify : with_iri? delete_clause insert_clause using_clause*  WHERE group_graph_pattern")
-def f(p): return _create_modify_query(p[0], p[1], p[2], p[3], p[5])
+@pg.production("modify : with_iri? insert_clause using_clause*  WHERE group_graph_pattern solution_modifier")
+def f(p): return _create_modify_query(p[0], [], p[1], p[2], p[4], p[5])
+@pg.production("modify : with_iri? delete_clause using_clause*  WHERE group_graph_pattern solution_modifier")
+def f(p): return _create_modify_query(p[0], p[1], [], p[2], p[4], p[5])
+@pg.production("modify : with_iri? delete_clause insert_clause using_clause*  WHERE group_graph_pattern solution_modifier")
+def f(p): return _create_modify_query(p[0], p[1], p[2], p[3], p[5], p[6])
 
 #@pg.production("delete_clause : DELETE quad_pattern")
 #@pg.production("insert_clause : INSERT quad_pattern")
@@ -398,19 +401,39 @@ def f(p):
   return r
 
 @pg.production("group_graph_pattern : { group_graph_pattern_item* }")
-def f(p):
+def _create_simple_block(p, accept_static = True):
   r = SimpleTripleBlock()
   for i in p[1]:
-    if   isinstance(i, Block) or isinstance(i, Bind) or isinstance(i, Filter):
-      r.append(i)
+    if   isinstance(i, Block) \
+      or isinstance(i, Bind) \
+      or isinstance(i, Filter):       r.append(i)
     elif isinstance(i, list):
-      r.extend(i)
-    else:
-      raise ValueError
+      in_static = False
+      for j in i:
+        if in_static:
+          static_triples.append(j)
+          if j.end_sequence:
+            _finalize_static(r, static_triples)
+            in_static = False
+        else:
+          if accept_static and isinstance(j, Triple) and (j[1].modifier == "*STATIC"):
+            j[1].modifier = "*"
+            static_triples = [j]
+            in_static = True
+          else: r.append(j)
+      if in_static: _finalize_static(r, static_triples)
+          
+    elif isinstance(i, StaticValues): r.static_valuess.append(i)
+    else: raise ValueError
   if (len(r) == 1) and isinstance(r[0], Block): return r[0]
   return r
 pg.list("group_graph_pattern+", "UNION")
 
+def _finalize_static(r, static_triples):
+  s = _create_simple_block(["{", [static_triples], "}"], accept_static = False)
+  static_block = StaticBlock(s)
+  r.static_valuess.append(static_block)
+  
 #@pg.production("group_graph_pattern_item_triple_content : triples_same_subject_path+")
 #def f(p): return p[0]
 #@pg.production("group_graph_pattern_item_triple_content : FILTER constraint")
@@ -447,19 +470,30 @@ def f(p):
 def f(p): return Bind(p[2], p[4])
 #@pg.production("group_graph_pattern_item : GRAPH var_or_iri group_graph_pattern")
 #@pg.production("group_graph_pattern_item : SERVICE SILENT? var_or_iri group_graph_pattern")
-#@pg.production("group_graph_pattern_item : inline_data")
+
+@pg.production("group_graph_pattern_item : inline_data")
+def f(p): return p[0]
+
+@pg.production("group_graph_pattern_item : STATIC group_graph_pattern")
+def f(p): return StaticBlock(p[1])
+
 pg.list("group_graph_pattern_item*", ".?")
 pg.optional(".?")
 
-#@pg.production("inline_data : VALUES data_block")
-#def f(p): return p
+@pg.production("inline_data : VALUES data_block")
+def f(p): return p[1]
 
-#@pg.production("data_block : inline_data_one_var")
-#@pg.production("data_block : inline_data_full")
-#def f(p): return p[0]
+@pg.production("data_block : inline_data_one_var")
+def f(p): return p[0]
+@pg.production("data_block : inline_data_full")
+def f(p): return p[0]
 
-#@pg.production("inline_data_one_var : var { data_block_value* }")
-#def f(p): return p
+@pg.production("inline_data_one_var : var { data_block_value* }")
+def f(p):
+  static_values = StaticValues()
+  static_values.vars.append(p[0])
+  static_values.valuess.extend([i.sql] if hasattr(i, "sql") else [i.value]  for i in p[2])
+  return static_values
 
 #@pg.production("nil_or_var* : NIL")
 #@pg.production("nil_or_var* : var*")
@@ -469,16 +503,26 @@ pg.optional(".?")
 #def f(p): return p
 #pg.list("data_block_value*_or_nil*", "")
 
+@pg.production("data_block_values : ( data_block_value* )")
+def f(p): return p[1]
+pg.list("data_block_values*", "")
+
 #@pg.production("inline_data_full : nil_or_var* { data_block_value*_or_nil* }")
-#def f(p): return p
+@pg.production("inline_data_full : ( var* ) { data_block_values* }")
+def f(p):
+  static_values = StaticValues()
+  for var in p[1]: static_values.vars.append(var)
+  static_values.valuess.extend([i.sql if hasattr(i, "sql") else i.value for i in l] for l in p[4])
+  return static_values
+
+@pg.production("data_block_value : iri")
+@pg.production("data_block_value : rdf_literal")
+@pg.production("data_block_value : numeric_literal")
+@pg.production("data_block_value : BOOL")
+@pg.production("data_block_value : undef")
+def f(p): return p[0]
+pg.list("data_block_value*", "")
 
-#@pg.production("data_block_value : iri")
-#@pg.production("data_block_value : rdf_literal")
-#@pg.production("data_block_value : numeric_literal")
-#@pg.production("data_block_value : BOOL")
-#@pg.production("data_block_value : UNDEF")
-#def f(p): return p[0]
-#pg.list("data_block_value*", "")
 
 @pg.production("constraint : bracketted_expression")
 @pg.production("constraint : builtin_call")
@@ -520,6 +564,57 @@ pg.list("property_list_not_empty_part+", ";")
 def f(p): return p
 #pg.optional("property_list_not_empty?")
 
+def _expand_triple(triples, s, ps_os):
+  for p, o in ps_os:
+    #if p.modifier == "*STATIC":
+    #  triples0 = triples
+    #  triples  = []
+    #  b = SimpleTripleBlock()
+    #  static_group = StaticGroup(b)
+    
+    if   isinstance(p, rply.Token):
+      if (p.name == "VAR") or (p.name == "PARAM"):
+        p.inversed = False
+        p.modifier = None
+      _add_triple(triples, s, p, o)
+      
+    elif isinstance(p, NegatedPropPath):
+      _add_triple(triples, s, p, o)
+      
+    elif isinstance(p, UnionPropPath):
+      if p.modifier:
+        _add_triple(triples, s, p, o)
+      else:
+        u = []
+        for p1 in p:
+          t = SimpleTripleBlock()
+          if p.inversed: _expand_triple(t, o, [(p1, s)])
+          else:          _expand_triple(t, s, [(p1, o)])
+          u.append(t)
+        triples.append(UnionBlock(u))
+        
+    elif isinstance(p, SequencePropPath):
+      if p.modifier:
+        raise NotImplementedError("Property path expressions of the form '(p1/p2)*' are not yet supported by Owlready2 native SPARQL engine. You may use RDFlib instead.")
+      else:
+        translator = CURRENT_TRANSLATOR.get()
+        s2 = s
+        for p2 in p:
+          if p2 is p[-1]: o2 = o
+          else:           o2 = rply.Token("VAR", translator.new_var())
+          #_add_triple(triples, s2, p2, o2)
+          _expand_triple(triples, s2, [(p2, o2)])
+          s2 = o2
+        triples[-1].end_sequence = True
+        
+    else: raise ValueError(p)
+    
+def _add_triple(triples, s, pr, o):
+  if isinstance(s, list): s = _expand_blank(triples, s)
+  if isinstance(o, list): o = _expand_blank(triples, o)
+  if getattr(pr, "inversed", False): triples.append(Triple([o, pr, s]))
+  else:                              triples.append(Triple([s, pr, o]))
+  
 def _expand_blank(triples, x):
   translator = CURRENT_TRANSLATOR.get()
   v = rply.Token("VAR", translator.new_var())
@@ -528,37 +623,6 @@ def _expand_blank(triples, x):
   _expand_triple(triples, v, x)
   return v
   
-def _add_triple(triples, s, pr, o):
-  if isinstance(s, list): s = _expand_blank(triples, s)
-  if isinstance(o, list): o = _expand_blank(triples, o)
-  if getattr(pr, "inversed", False): triples.append(Triple([o, pr, s]))
-  else:                              triples.append(Triple([s, pr, o]))
-
-def _expand_triple(triples, s, ps_os):
-  for p, o in ps_os:
-    if   isinstance(p, rply.Token) and ((p.name == "VAR") or (p.name == "PARAM")):
-      p.inversed = False
-      p.modifier = None
-      _add_triple(triples, s, p, o)
-    elif len(p) > 1: # '|' in path
-      u = []
-      for p1 in p:
-        t = SimpleTripleBlock()
-        _expand_triple(t, s, [([p1], o)])
-        u.append(t)
-      triples.append(UnionBlock(u))
-    else:
-      if len(p[0]) == 1:
-        _add_triple(triples, s, p[0][0], o)
-      else: # '/' in path
-        translator = CURRENT_TRANSLATOR.get()
-        s2 = s
-        for p2 in p[0]:
-          if p2 is p[0][-1]: o2 = o
-          else:              o2 = rply.Token("VAR", translator.new_var())
-          _add_triple(triples, s2, p2, o2)
-          s2 = o2
-  
   
 @pg.production("triples_same_subject_path : var_or_term property_list_path_not_empty")
 @pg.production("triples_same_subject_path : triples_node_path property_list_path_not_empty?")
@@ -589,10 +653,16 @@ def f(p): # Expand ',' in triples
 pg.optional("property_list_path_not_empty?")
 
 @pg.production("path : path_sequence+")
-def f(p): return p[0]
+def f(p):
+  p = p[0]
+  if len(p) == 1: return p[0]
+  return UnionPropPath(p)
 
 @pg.production("path_sequence : path_element_or_inverse+")
-def f(p): return p[0]
+def f(p):
+  p = p[0]
+  if len(p) == 1: return p[0]
+  return SequencePropPath(p)
 pg.list("path_sequence+", "|")
 
 @pg.production("path_element_or_inverse : path_element")
@@ -600,18 +670,17 @@ def f(p): return p[0]
 @pg.production("path_element_or_inverse : ^ path_element")
 def f(p):
   p = p[1]
-  if isinstance(p, list): p = _List(p)
   p.inversed = True
   return p
 pg.list("path_element_or_inverse+", "/")
 
 @pg.production("path_element : path_primary path_mod?")
 def f(p):
-  if isinstance(p[0], list): p[0] = _List(p[0])
   p[0].modifier = p[1] and p[1].value
   return p[0]
 
 @pg.production("path_mod : ?")
+@pg.production("path_mod : *STATIC")
 @pg.production("path_mod : *")
 @pg.production("path_mod : +")
 def f(p): return p[0]
@@ -624,11 +693,11 @@ def f(p):
   return p[0]
 @pg.production("path_primary : ! path_negated_property_set")
 def f(p):
-  prop_path = NegatedPropSetPath()
-  for i in p[1]: prop_path.props.append(i)
+  prop_path = NegatedPropPath()
+  for i in p[1]: prop_path.append(i)
   return prop_path
 @pg.production("path_primary : ( path )")
-def f(p): return p
+def f(p): return p[1]
 
 @pg.production("path_negated_property_set : path_one_in_property_set")
 def f(p): return p
@@ -645,6 +714,7 @@ def f(p):
   return p[0]
 pg.list("path_one_in_property_set+", "|")
 
+
 @pg.production("triples_node : collection")
 @pg.production("triples_node : blank_node_property_list")
 def f(p): return p[0]
@@ -710,16 +780,17 @@ def f(p): return p[0]
 @pg.production("param : PARAM")
 def f(p):
   p = p[0]
+  translator = CURRENT_TRANSLATOR.get()
   if len(p.value) > 2:
     p.number = int(p.value[2:])
+    if p.number > translator.max_fixed_parameter: translator.max_fixed_parameter = p.number
   else:
-    translator = CURRENT_TRANSLATOR.get()
     p.number = translator.new_parameter()
   return p
 
 @pg.production("conditional_or_expression_operand : || conditional_and_expression")
 def f(p):
-  p[0].sql = "OR"
+  p[0].sql = "OR "
   return p
 pg.list("conditional_or_expression_operand*", "")
 @pg.production("expression : conditional_and_expression conditional_or_expression_operand*")
@@ -729,7 +800,7 @@ pg.list("expression+", ",", keep_sep = True)
 
 @pg.production("conditional_and_expression_operand : && relational_expression")
 def f(p):
-  p[0].sql = "AND"
+  p[0].sql = "AND "
   return p
 pg.list("conditional_and_expression_operand*", "")
 @pg.production("conditional_and_expression : relational_expression conditional_and_expression_operand*")
@@ -769,6 +840,9 @@ def f(p):
   return r
 
 @pg.production("unary_expression : ! primary_expression")
+def f(p):
+  p[0].sql = "NOT "
+  return p
 @pg.production("unary_expression : + primary_expression")
 @pg.production("unary_expression : - primary_expression")
 def f(p): return p
@@ -797,6 +871,11 @@ def f(p): return p[0]
 @pg.production("builtin_call : aggregate")
 def f(p): return p[0]
 
+@pg.production("func : iri ( expression* )")
+def f(p):
+  func_name = p[0].value.upper()
+  p[0].name = "FUNC"
+  return p
 @pg.production("func : FUNC ( expression* )")
 @pg.production("func : FUNC NIL")
 def f(p):
@@ -887,6 +966,13 @@ def f(p):
   p.value = "'%s'" % p.value[3:-3].replace("'", r"\'")
   return p
 
+@pg.production("unabbreviated_iri : IRI")
+def f(p):
+  translator = CURRENT_TRANSLATOR.get()
+  p = p[0]
+  p.value  = p.value[1:-1]
+  if not ":" in p.value: p.value = "%s%s" % (translator.base_iri, p.value) # Relative IRI
+  return p
 @pg.production("iri : IRI")
 def f(p):
   translator = CURRENT_TRANSLATOR.get()
@@ -924,6 +1010,12 @@ def f(p):
   p.value  = CURRENT_TRANSLATOR.get().new_var()
   return p
 
+@pg.production("undef : UNDEF")
+def f(p):
+  p = p[0]
+  p.value = p.sql = "NULL"
+  return p
+
 PARSER = pg.build()
 del lg, pg
 
@@ -941,7 +1033,9 @@ def _prefix_vars(x, prefix):
     elif isinstance(i, rply.Token) and i.name == "VAR": i.value = "%s%s" % (prefix, i.value)
     
 class Block(list):
-  def __repr__(self): return "<%s %s>" % (self.__class__.__name__, list.__repr__(self))
+  def __repr__(self):
+    if getattr(self, "static_valuess"): return "<%s %s static=%s>" % (self.__class__.__name__, list.__repr__(self), self.static_valuess)
+    return "<%s %s>" % (self.__class__.__name__, list.__repr__(self))
 
   def get_ordered_vars(self):
     vars = set()
@@ -959,27 +1053,39 @@ class UnionBlock(Block):
     
     self.simple_union_triples = self._to_simple_union()
     
+  def __repr__(self): return "<%s %s %s>" % (self.__class__.__name__, "Simple" if self.simple_union_triples else "Non-Simple", list.__repr__(self))
+  
   def _to_simple_union(self):
     if len(self[0]) == 2:
       r = self._to_simple_union2()
       if r: return r
       
     for i in self:
+      if not isinstance(i, SimpleTripleBlock): return None
       if len(i) > 1: return None
+      if (not i): return None
+      if (not isinstance(i[0], Triple)): return None
       
     ss = set()
     ps = set()
     os = set()
+    mods = set()
     for i in self:
       ss.add(repr(i[0][0]))
       ps.add(repr(i[0][1]))
       os.add(repr(i[0][2]))
+      mods.add(i[0][1].modifier)
+    if len(mods) > 1: return None
     nb_many = 0
     if len(ss) > 1: nb_many += 1; n = 0
     if len(ps) > 1: nb_many += 1; n = 1
     if len(os) > 1: nb_many += 1; n = 2
-    if nb_many > 1: return None
-    
+    if nb_many > 1:  return None
+    if nb_many == 0: return None
+    if n == 1: # Prop; prop path modifier not supported in that case
+      for i in self:
+        if i[0][1].modifier: return None
+        
     vs = [i[0][n] for i in self]
     for v in vs:
       if (not isinstance(v, rply.Token)) or (v.name != "IRI"): return None
@@ -1023,9 +1129,17 @@ def _var_found(var, vars, ordered_vars):
     ordered_vars.append(var)
     
 class SimpleTripleBlock(TripleBlock):
+  def __init__(self, l = None):
+    TripleBlock.__init__(self, l or [])
+    
+    self.static_valuess = []
+    
   def _get_ordered_vars(self, vars, ordered_vars, root_call = False):
     for triple in self:
       triple._get_ordered_vars(vars, ordered_vars)
+
+    for static in self.static_valuess:
+      static._get_ordered_vars(vars, ordered_vars)
       
 class OptionalBlock(TripleBlock):
   def _get_ordered_vars(self, vars, ordered_vars, root_call = False):
@@ -1048,9 +1162,12 @@ class SubQueryBlock(Block):
 
 
 class Triple(tuple):
+  end_sequence = False
   def __init__(self, l):
-    self.optional = False
-    self.var_names = { x.value for x in self if x.name == "VAR" }
+    self.optional     = False
+    self.likelihood_p = None
+    self.likelihood_o = None
+    self.var_names    = { x.value for x in self if x.name == "VAR" }
     p_storid = getattr(self[1], "storid", None)
     self.Prop = p_storid and CURRENT_TRANSLATOR.get().world._get_by_storid(p_storid)
     
@@ -1062,12 +1179,14 @@ class Triple(tuple):
     elif isinstance(self.Prop, ObjectPropertyClass): self.table_type = "objs"
     elif isinstance(self.Prop, DataPropertyClass):   self.table_type = "datas"
     else:                                            self.table_type = "quads"
+      
     
   def _get_ordered_vars(self, vars, ordered_vars):
     if self[1].inversed: triple = self[::-1]
     else:                triple = self
     for x in triple:
-      if x.name == "VAR": _var_found(x.value, vars, ordered_vars)
+      if   x.name == "VAR": _var_found(x.value, vars, ordered_vars)
+      elif isinstance(x, PropPath): x._get_ordered_vars(vars, ordered_vars)
       
   
 class SpecialCondition(object): pass
@@ -1084,24 +1203,42 @@ class SimpleUnion(SpecialCondition):
 
   def __repr__(self): return "<SimpleUnion %s>" % self.items
   def __str__ (self): return ",".join(str(i) for i in self.items)
-    
-class PropPath(SpecialCondition): pass
-class NegatedPropSetPath(PropPath):
-  name = "NEGATED_PROP_PATH"
-  def __init__(self):
-    self.props = []
-
-  def create_conditions(self, conditions, table, n):
-    if len(self.props) == 1:
-      conditions.append("%s.%s!=%s" % (table.name, n, self.props[0].storid))
-    else:
-      conditions.append("%s.%s NOT IN (%s)" % (table.name, n, ",".join(str(p.storid) for p in self.props)))
+  
+  
+class PropPath(list):
+  def __init__(self, l = None):
+    list.__init__(self, l or [])
+    self.inversed = False
+    self.modifier = None
+  __hash__ = object.__hash__
+  
+  def _get_ordered_vars(self, vars, ordered_vars):
+    for x in self:
+      if   x.name == "VAR": _var_found(x.value, vars, ordered_vars)
+      elif isinstance(x, PropPath): x._get_ordered_vars(vars, ordered_vars)
       
+class NegatedPropPath(PropPath, SpecialCondition):
+  name = "NEGATED_PROP_PATH"
+  def create_conditions(self, conditions, table, n):
+    if len(self) == 1:
+      conditions.append("%s.%s!=%s" % (table.name, n, self[0].storid))
+    else:
+      conditions.append("%s.%s NOT IN (%s)" % (table.name, n, ",".join(str(p.storid) for p in self)))
+  def __repr__(self): return "!(%s)" % " ".join(repr(prop) for prop in self)
+      
+class SequencePropPath(PropPath):
+  name = "SEQUENCE_PROP_PATH"
+  def __repr__(self): return "(%s)" % "/".join(repr(prop) for prop in self)
+  
+class UnionPropPath(PropPath):
+  name = "UNION_PROP_PATH"
+  def __repr__(self): return "(%s)" % "|".join(repr(prop) for prop in self)
+  
 
 def _get_vars(x):
   for i in x:
     if   isinstance(i, rply.Token) and i.name == "VAR": yield i.value
-    elif isinstance(i, list): yield from _get_vars(i)
+    elif isinstance(i, (list, tuple)): yield from _get_vars(i)
   
 class Bind(object):
   def __init__(self, expression, var):
@@ -1127,7 +1264,40 @@ class Filter(object):
     self.constraint = constraint
     self.var_names = set(_get_vars(constraint))
     
-  def _get_ordered_vars(self, vars, ordered_vars): pass
+  def _get_ordered_vars(self, vars, ordered_vars):
+    for var in self.var_names:
+      _var_found(var, vars, ordered_vars)
+    
   
+class StaticValues(object):
+  def __init__(self):
+    self.vars    = []
+    self.valuess = []
+    
+  def _get_ordered_vars(self, vars, ordered_vars, root_call = False): pass
+  
+  def __repr__(self): return "VALUES (%s) %s" % (" ".join(repr(var) for var in self.vars), self.valuess)
   
   
+class StaticBlock(StaticValues):
+  def __init__(self, blocks):
+    StaticValues.__init__(self)
+    self.inner_blocks = blocks
+    self.ordered_vars = list(_get_vars(blocks))
+    self.all_vars     = set(_get_vars(blocks))
+    
+    self.old_translator = CURRENT_TRANSLATOR.get()
+    self.translator = self.old_translator.make_translator()
+    CURRENT_TRANSLATOR.set(self.translator)
+    
+    self.translator.main_query = self.translator.new_sql_query("main", self.inner_blocks, None, True)
+    
+    self.translator.main_query.type = "select"
+    CURRENT_TRANSLATOR.set(self.old_translator)
+    del self.old_translator
+    
+  def _get_ordered_vars(self, vars, ordered_vars, root_call = False):
+    for var in self.ordered_vars: _var_found(var, vars, ordered_vars)
+    
+    
+  def __repr__(self): return """<StaticGraph vars=%s>""" % ",".join(repr(var) for var in self.all_vars)
